"""
Particles.
"""

import os

import numpy as np

from ..cpp import Types
from ..utils import strnum
from .effect import Effect


class Particles(Effect):
    """
    Particles emit from keys.
    """

    def render(self, props, img: np.ndarray, frame: int, notes):
        """
        Render particles.

        :param notes: MIDI notes.
        """
        keys = np.array([n.note for n in notes], dtype=Types.int)
        starts = np.array([n.start for n in notes], dtype=Types.double)
        ends = np.array([n.end for n in notes], dtype=Types.double)

        cache_in = self.cache / "ptcls" / strnum(frame-1)
        cache_out = self.cache / "ptcls" / strnum(frame)
        if not cache_in.is_file():
            cache_in = ""

        props_args = [props.video.fps, props.ptcls.pps, props.ptcls.air_resist,
            props.ptcls.lifetime, props.ptcls.x_vel, props.ptcls.y_vel,
            props.ptcls.wind_strength, props.ptcls.heat_strength, props.ptcls.gravity]

        self.libs["ptcls"].render_ptcls(
            img, img.shape[1], img.shape[0],
            frame,
            Types.cpath(cache_in), Types.cpath(cache_out),
            len(notes), keys, starts, ends,
            *props_args,
        )



"""
Rendering pipeline.
"""

import ctypes
import json
import os
from pathlib import Path
from typing import Mapping

import cv2
import numpy as np
from tqdm import trange

from .. import logger
from ..effects import parse_midi
from ..effects import Blocks, Keyboard, Glare, Particles
from ..utils import bounds
from .lib import load_libs
from .video import Video


def check_previous(args, settings, cache):
    """
    Check if continue previous render.

    :return: Frame to start rendering from.
    """
    cache_settings = cache / "settings.json"
    cache_curr = cache / "currently_rendering.txt"

    real_start = None
    if cache_settings.is_file() and cache_curr.is_file():
        with open(cache_settings, "r") as fp:
            prev_sets = Settings(json.load(fp))  # TODO change

        if prev_sets == settings:
            with open(cache_curr, "r") as fp:
                data = fp.read()
                real_start = int(data) if data.isdigit() else None

            if real_start is not None:
                if args.resume is None:
                    print("Last render has same settings as this render, "
                          f"and stopped at frame {real_start}.")
                    if input("Continue last render? [Y/n] ").lower().strip() \
                            == "n":
                        real_start = None

                elif not args.resume:
                    real_start = None

    return real_start


def render_video(args, scene, out: str, cache: Path) -> None:
    """
    Render system main.

    :param args: Argparse arguments.
    """
    # Libraries
    try:
        libs = load_libs(cache)
    except AssertionError:
        logger.error("Failed to build libraries.")
        raise

    # Cache subdirs
    for sub in ("glare", "ptcls"):
        (cache/sub).mkdir(exist_ok=True)

    # Save settings to cache.
    """
    os.makedirs(cache, exist_ok=True)
    real_start = check_previous(args, settings, cache)
    with open(cache / "settings.json", "w") as fp:
        json.dump(settings._json(), fp)
    """
    real_start = None  # Remove later

    props = scene.default
    video = Video(cache / "output")
    render_frames(scene, libs, video, cache, real_start)
    video.compile(out, props)


def get_frame_bounds(props, duration):
    """
    Returns (start_frame, end_frame).

    :param duration: Duration of MIDI in frames.
    """
    fps = props.video.fps
    m_start = props.composition.margin_start
    m_end = props.composition.margin_end
    frame_start = -fps * m_start
    frame_end = duration + fps*m_end

    return (frame_start, frame_end)


def add_fade(props, img, frame_start, frame_end, frame):
    fps = props.video.fps
    fade_in = frame_start + props.composition.fade_in * fps
    fade_out = frame_end - props.composition.fade_out * fps
    fade_blur = props.composition.fade_blur

    fade_fac = 1
    if frame <= fade_in:
        fade_fac *= np.interp(frame, (frame_start, fade_in), (0, 1))
    if frame >= fade_out:
        fade_fac *= np.interp(frame, (fade_out, frame_end), (1, 0))
    fade_fac = bounds(fade_fac, 0, 1)

    if fade_fac < 1:
        blur = int(fade_blur * (1-fade_fac))
        img[...] = (img * fade_fac).astype(np.uint8)
        if blur > 0:
            img[...] = cv2.blur(img, (blur, blur))


def render_frames(scene, libs, video, cache, real_start=None) -> int:
    """
    Render frames.

    :param real_start: Frame to start rendering from.
    :return: Number of frames rendered.
    """
    # Parse MIDI.
    notes = parse_midi(scene.default)
    duration = int(max(x.end for x in notes))

    # Calculate start and end.
    frame_start, frame_end = get_frame_bounds(scene.default, duration)

    # OOP effects
    props = scene.default
    blocks = Blocks(props, cache, libs)
    keyboard = Keyboard(props, cache, libs, notes)
    glare = Glare(props, cache, libs, notes)
    ptcls = Particles(props, cache, libs)

    # Adjust start to match previous render
    if real_start is None:
        real_start = frame_start
    frame_start, frame_end, real_start = map(int,
        (frame_start, frame_end, real_start))
    logger.info(f"Starting render from frame {real_start}")

    # Render
    num_frames = real_start - frame_start
    video.frame = num_frames
    for frame in trange(real_start, frame_end, desc="Rendering"):
        # Save state
        num_frames += 1
        with open(cache/"currently_rendering.txt", "w") as fp:
            fp.write(str(frame))

        # Create image
        img = np.zeros((*scene.default.video.resolution[::-1], 3),
            dtype=np.uint8)

        # Apply effects
        props = scene.values(frame)
        keyboard.render(props, img, frame)
        blocks.render(props, img, frame, notes)
        ptcls.render(props, img, frame, notes)
        glare.render(props, img, frame, notes)

        # Fade
        add_fade(scene.default, img, frame_start, frame_end, frame)

        video.write(img)



from typing import Any

from .accessor import Accessor


class Modifier:
    """
    Base modifier class. A modifier instance can be called,
    and modifies a value. Examples include multiplying by 0.5
    to dim colors (a Dim modifier).

    Base class modifier does nothing.
    """

    def __init__(self):
        pass

    def __call__(self, default: Accessor, value: Any) -> Any:
        """
        Modify the value.

        :param default: Default values, if you need any.
        :param value: The value to modify.
        """
        return value


class Coords(Modifier):
    """
    Modifier from coords value to pixel value.
    """

    def __call__(self, default, value):
        coord = default.video.resolution[0] / 52
        return value * coord


class SecToFrame(Modifier):
    """
    Convert seconds to frames.
    """

    def __call__(self, default, value):
        return value / default.video.fps



"""
Global utilities.
"""

import os
import shutil
from pathlib import Path

from . import logger

VERSION = "0.2.3"

ROOT = Path(__file__).absolute().parent

FFMPEG = shutil.which("ffmpeg")
GCC = shutil.which("g++")
if FFMPEG is None:
    logger.warn("No FFmpeg.")
if GCC is None:
    logger.warn("No g++.")


def forever():
    """
    Generator for indefinite tqdm progressbar.
    """
    while True:
        yield


def strnum(v):
    """
    Returns str(v) if v >= 0 else "n" + str(abs(v))
    Used for getting a filename given frame number.
    """
    return str(v) if v >= 0 else "n" + str(abs(v))


def bounds(v, vmin, vmax):
    """
    Bound v between vmin and vmax.
    """
    return min(max(v, vmin), vmax)

def interp(v, from_range, to_range):
    """
    Like np.interp, except better.
    Can handle v out of from_range, reversed ranges, etc.
    """
    fac = (v-from_range[0]) / (from_range[1]-from_range[0])
    return fac * (to_range[1]-to_range[0]) + to_range[0]



import os
import random

import numpy as np

from ..cpp import Types
from .effect import Effect


class Glare(Effect):
    """
    Light glare when notes play.
    """

    def __init__(self, props, cache, libs, notes) -> None:
        super().__init__(props, cache, libs)

        for note in notes:
            streaks = []
            for _ in range(props.glare.streaks):
                angle = random.randint(0, 255)
                streaks.append(angle)
            note.attrs["glare.streak_angles"] = streaks

    def render(self, props, img: np.ndarray, frame: int, notes):
        """
        Render the glare.

        :param notes: MIDI notes from parse_midi.
        """
        keys = np.array([n.note for n in notes], dtype=Types.int)
        starts = np.array([n.start for n in notes], dtype=Types.double)
        ends = np.array([n.end for n in notes], dtype=Types.double)

        angles = []
        for note in notes:
            angles.extend(note.attrs["glare.streak_angles"])
        angles = np.array(angles, dtype=np.uint8)

        props_args = [props.piano.black_width_fac,
            props.glare.radius, props.glare.intensity,
            props.glare.jitter, props.glare.streaks]

        self.libs["glare"].render_glare(
            img, img.shape[1], img.shape[0],
            frame,
            len(notes), keys, starts, ends, angles,
            *props_args,
        )



import numpy as np
import pygame

pygame.init()

TMP = "/tmp"

BLACK = np.array((0, 0, 0))
DARK_GRAY = np.array((50, 50, 50))
GRAY = np.array((120, 120, 120))
LIGHT_GRAY = np.array((170, 170, 170))
WHITE = np.array((255, 255, 255))
BLUE = np.array((130, 130, 255))

FONT_MED = pygame.font.SysFont("ubuntu", 18)



import cv2
import pygame

from .timeline import Timeline
from .utils import *
from .video import Video

pygame.init()


class Preview:
    """
    Preview current frame.
    """

    def __init__(self, video: Video, timeline: Timeline):
        self.video = video
        self.timeline = timeline

    def draw(self, surface: pygame.Surface, events, rect):
        """
        Draw info on surface.

        :param rect: ``(x, y, w, h)`` coordinates to draw.
        """
        x, y, w, h = rect

        if self.timeline.frame > self.video.extracted:
            return

        img = self.video.get(self.timeline.frame)
        img = cv2.resize(img, (w, h)).swapaxes(0, 1)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = pygame.surfarray.make_surface(img)
        surface.blit(img, (x, y))



"""
Main entry point.
"""

import argparse
import json
import os
import sys
from ast import literal_eval
from pathlib import Path

from . import logger
from .api import import_scene
from .utils import VERSION

from .render import render_video
from .view import view_video


def render(args):
    """
    Call this when the user requests render, e.g.
    pianoray render ...
    """
    if args.output.exists() and not args.yes:
        if input(f"Overwrite output file {args.output}? [y/N] ") \
                .lower().strip() != "y":
            return 3

    logger.info(f"Using scene {args.clsname} from {args.file}")
    scene = import_scene(args.file, args.clsname)()
    render_video(args, scene, args.output, args.cache)

    if args.preview:
        logger.info("Opening output with xdg-open")
        os.system(f"xdg-open {args.output}")


def view(args):
    """
    Called when the user requests view e.g.
    pianoray view ...
    """
    view_video(args.file)


def main():
    """
    Main entry point.
    """
    version_str = f"Pianoray v{VERSION}"

    parser = argparse.ArgumentParser(
        description="Piano performance visualizer.")
    subparsers = parser.add_subparsers(title="subcommands", dest="subparser")
    parser.add_argument("-V", "--version", action="version",
        help="Show version info.", version=version_str)
    parser.add_argument("-y", "--yes",
        help="Don't prompt overwrite.", action="store_true")

    render_parser = subparsers.add_parser("render",
        help="Render a video.")
    render_parser.add_argument("file",
        help="Path to Python file containing scene.")
    render_parser.add_argument("clsname",
        help="Name of the scene class object.")
    render_parser.add_argument("-o", "--output", required=True,
        help="Output video file.", type=Path)
    render_parser.add_argument("-c", "--cache", default=".prcache",
        help="Cache path (default .prcache)", type=Path)
    render_parser.add_argument("-p", "--preview", action="store_true",
        help="Open output file after rendering")
    render_parser.add_argument("-r", "--resume", type=literal_eval,
        help="Whether to resume previous render (omit for prompt).")

    view_parser = subparsers.add_parser("view",
        help="View a video file in a GUI video.")
    view_parser.add_argument("file", help="Path to file to view.")

    args = parser.parse_args()
    logger.info(version_str)

    if args.subparser == "render":
        render(args)
    elif args.subparser == "view":
        view(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    sys.exit(main())



"""
Video writer.
"""

import os
import shutil
from pathlib import Path
from subprocess import Popen, PIPE
from typing import Sequence

import cv2
import numpy as np

from .. import logger
from ..utils import FFMPEG


class Video:
    """
    Video.

    Saves frames to given cache directory and uses ffmpeg to make the
    final video.
    """

    def __init__(self, cache: Path) -> None:
        """
        Initializes video.

        :param cache: Cache directory. Frames stored there.
        """
        self.cache = cache
        self.frame = 0

        (self.cache/"frames").mkdir(parents=True, exist_ok=True)

    def write(self, img: np.ndarray) -> int:
        """
        Write a frame.
        Converts from RGB to BGR.

        :param img: Frame of shape ``(height, width, 3)``
        :return: This frame number.
        """
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        path = str(self.cache / "frames" / f"{self.frame}.jpg")
        cv2.imwrite(path, img)

        self.frame += 1
        return self.frame - 1

    def compile(self, out: str, props) -> None:
        """
        Use ffmpeg to compile frames and audio to video.

        :param out: Output video path.
        """
        # Frames to video
        logger.info("Compiling frames to video.")
        args = [
            FFMPEG,
            "-y",
            "-r", props.video.fps,
            "-start_number", 0,
            "-i", self.cache / "frames" / "%d.jpg",
            "-vframes", self.frame,
            "-c:v", props.video.vcodec, "-an",
            "-crf", 24,
            "-r", props.video.fps,
            self.cache / "no_audio.mp4",
        ]
        run_ffmpeg(args)

        if props.audio.file is not None:
            # Cut audio
            logger.info("Processing audio.")
            args = [
                FFMPEG,
                "-y",
                "-ss", props.audio.start - props.composition.margin_start,
                "-t", self.frame / props.video.fps,
                "-i", props.audio.file,
                self.cache / "offset.mp3",
            ]
            run_ffmpeg(args)

            # Mix
            logger.info("Combining audio and video.")
            args = [
                FFMPEG,
                "-y",
                "-i", self.cache / "no_audio.mp4",
                "-i", self.cache / "offset.mp3",
                "-c", "copy",
                out,
            ]

            run_ffmpeg(args)

        else:
            # Copy to output.
            shutil.copy(self.cache / "no_audio.mp4", out)

        logger.info(f"Video saved to {out}")


def run_ffmpeg(args: Sequence[str]):
    args = list(map(str, args))
    proc = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    proc.wait()

    if proc.returncode != 0:
        msg = f"FFmpeg exited with code {proc.returncode}. " + \
            "Command: " + " ".join(args)
        raise ValueError(msg)



import shutil
import time

import pygame

from .. import logger
from .info import Info
from .preview import Preview
from .timeline import Timeline
from .utils import *
from .video import Video

pygame.init()


def view_video(path: str) -> None:
    logger.warn("Viewer is not complete yet.")

    video = Video(path)
    timeline = Timeline(video)
    info = Info(video, timeline)
    preview = Preview(video, timeline)

    logger.info(f"Extracting frames to {video.tmpdir}")

    resized = False  # Redraw if resized
    display = pygame.display.set_mode((1280, 720), pygame.RESIZABLE)
    pygame.display.set_caption("PianoRay Viewer")

    try:
        run = True
        while run:
            time.sleep(1/15)
            resized = False

            pygame.display.update()
            events = pygame.event.get()
            for event in events:
                if event.type == pygame.QUIT:
                    run = False

                elif event.type == pygame.VIDEORESIZE:
                    resized = True

                elif event.type == pygame.KEYDOWN:
                    if event.key in (pygame.K_RIGHT, pygame.K_l):
                        timeline.next_frame()
                    elif event.key in (pygame.K_LEFT, pygame.K_h):
                        timeline.prev_frame()

            width, height = display.get_size()
            v_split = int(height * 0.9)
            h_split = int(width * 0.8)

            display.fill(BLACK)
            timeline.draw(display, events, (0, v_split, width, height-v_split))
            info.draw(display, events, (h_split, 0, width-h_split, v_split))
            preview.draw(display, events, (0, 0, h_split, v_split))
            pygame.draw.line(display, WHITE, (0, v_split), (width, v_split))
            pygame.draw.line(display, WHITE, (h_split, 0), (h_split, v_split))

    finally:
        video.run = False
        time.sleep(0.05)  # Wait for thread to stop

        pygame.quit()
        shutil.rmtree(video.tmpdir)



"""
Handles loading libraries.
Calls functions from ../cpp.py
"""

import ctypes
import os
from pathlib import Path
from typing import Mapping

from ..cpp import build_lib, Types


def load_libs(cache: Path) -> Mapping[str, ctypes.CDLL]:
    """
    Load C libraries. Also sets the argtypes and restypes.
    """
    cache = cache / "c_libs"
    cache.mkdir(parents=True, exist_ok=True)

    blocks = build_lib(
        ["cutils/blocks.cpp"],
        cache,
        "blocks",
    )
    blocks.render_blocks.argtypes = [
        Types.img, Types.int, Types.int,
        Types.int,
        Types.int, Types.arr_int, Types.arr_double, Types.arr_double,
        Types.int, Types.double, Types.double, Types.double, Types.arr_uchar,
            Types.double, Types.double, Types.arr_uchar,
    ]

    glare = build_lib(
        ["cutils/glare.cpp"],
        cache,
        "glare",
    )
    glare.render_glare.argtypes = [
        Types.img, Types.int, Types.int,
        Types.int,
        Types.int, Types.arr_int, Types.arr_double, Types.arr_double,
            Types.arr_uchar,
        Types.double, Types.double, Types.double, Types.double, Types.int,
    ]

    keyboard = build_lib(
        ["cutils/keyboard.cpp"],
        cache,
        "keyboard",
    )
    keyboard.render_octave_lines.argtypes = [
        Types.img, Types.int, Types.int,
    ]

    particles = build_lib(
        ["cutils/particles.cpp"],
        cache,
        "particles",
    )
    particles.render_ptcls.argtypes = [
        Types.img, Types.int, Types.int,
        Types.int,
        Types.path, Types.path,
        Types.int, Types.arr_int, Types.arr_double, Types.arr_double,
        Types.int, Types.double, Types.double, Types.double, Types.double, Types.double,
            Types.double, Types.double, Types.double,
    ]

    return {
        "blocks": blocks,
        "glare": glare,
        "keyboard": keyboard,
        "ptcls": particles,
    }



from typing import Any


class Interp:
    """
    Namespace for interpolation types.
    """
    CONSTANT = "CONSTANT"
    LINEAR = "LINEAR"
    QUADRATIC = "QUADRATIC"


class Keyframe:
    """
    Animate a property.
    """
    frame: int
    value: Any
    interp: str

    def __init__(self, frame: int, value: Any, interp: str) -> None:
        self.frame = frame
        self.value = value
        self.interp = interp

    def __eq__(self, other):
        return isinstance(other, Keyframe) and self.frame == other.frame

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return isinstance(other, Keyframe) and self.frame < other.frame

    def __gt__(self, other):
        return isinstance(other, Keyframe) and self.frame > other.frame

    def __le__(self, other):
        return not self > other

    def __ge__(self, other):
        return not self < other

    def __repr__(self) -> str:
        return f"Keyframe(frame={self.frame}, value={self.value}, " + \
               f"interp={self.interp})"



"""
Blocks effect.
"""

import numpy as np

from ..cpp import Types
from .effect import Effect


class Blocks(Effect):
    """
    The blocks that fall down.
    """

    def render(self, props, img: np.ndarray, frame: int, notes):
        """
        Render the blocks.
        :param notes: MIDI notes from parse_midi.
        """
        keys = np.array([n.note for n in notes], dtype=Types.int)
        starts = np.array([n.start for n in notes], dtype=Types.double)
        ends = np.array([n.end for n in notes], dtype=Types.double)

        props_args = [props.video.fps, props.blocks.speed,
            props.piano.black_width_fac, props.blocks.radius,
            np.array(props.blocks.color, dtype=np.uint8),
            props.blocks.glow_intensity, props.blocks.glow_radius,
            np.array(props.blocks.glow_color, dtype=np.uint8),
        ]

        self.libs["blocks"].render_blocks(
            img, img.shape[1], img.shape[0],
            frame,
            len(notes), keys, starts, ends,
            *props_args,
        )



import ctypes
from pathlib import Path
from typing import Mapping

import numpy as np

from ..api.accessor import Accessor


class Effect:
    """
    Base class for all effects.
    """
    cache: Path
    libs: Mapping[str, ctypes.CDLL]

    def __init__(self, props: Accessor, cache: Path, libs) -> None:
        self.cache = cache
        self.libs = libs

    def render(self, props: Accessor, img: np.ndarray, frame: int,
            *args, **kwargs) -> None:
        """
        Override this in the subclass.
        """
        raise NotImplementedError("Override Effect.render")



import os
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Type

import numpy as np

from .accessor import Accessor
from .keyframe import Keyframe, Interp
from .interpolate import interpolate
from .modifiers import Modifier

__all__ = (
    "Property",
    "BoolProp",
    "IntProp",
    "FloatProp",
    "StrProp",
    "PathProp",
    "ArrayProp",
    "RGBProp",
)


class Property:
    """
    Property base class.
    """
    type: Type
    supported_interps: Iterable[str]

    name: str
    desc: str
    animatable: bool
    required: bool
    mods: Sequence[Modifier]
    default: Any

    _keyframes: List[Keyframe]
    _value: Any

    def __init__(self, name: str = "", desc: str = "", animatable: bool = True,
            required: bool = True, mods: Sequence[Modifier] = (),
            default: Optional[Any] = None):
        """
        Initialize property with common arguments for all subclasses.

        :param name: Human readable name of the property. Can be different from
            the variable name in Python.
        :param desc: Human readable description.
        :param animatable: Whether this property can be animated.
        :param required: If False, value can be None.
        :param mods: Sequence of modifiers to apply in order.
        :param default: Default value.
        """
        self.name = name
        self.desc = desc
        self.animatable = animatable
        self.required = required
        self.mods = mods
        self.default = default if default is None else self.type(default)

        self._keyframes = []
        self._value = None

        if self.default is not None:
            assert self.verify(self.default)

    def set_value(self, value: Any):
        """
        Checks validity and sets self._value
        """
        value = self.type(value)
        assert self.verify(value)
        self._value = value

    def animate(self, *args) -> None:
        """
        Insert a keyframe.

        A few syntaxes are available:

        .. code-block:: py

           prop.animate(Keyframe(frame, value, interp))
           prop.animate(Keyframe(frame, value, interp), Keyframe(frame, value, interp))
           prop.animate(frame, value, interp)
           prop.animate((frame, value, interp))
           prop.animate((frame, value, interp), (frame2, value2, interp2), ...)

        They all do the same thing. However, please do not mix syntaxes in one call
        (don't pass a keyframe object and then an unpacked tuple).
        """
        assert self.animatable

        keyframes = []
        if isinstance(args[0], (tuple, Keyframe)):
            for a in args:
                k = a if isinstance(a, Keyframe) else Keyframe(*a)
                keyframes.append(k)
        else:
            keyframes.append(Keyframe(*args))

        assert all(k.interp in self.supported_interps for k in keyframes)
        for k in keyframes:
            k.value = self.type(k.value)
        assert all(self.verify(k.value) for k in keyframes)
        self._keyframes.extend(keyframes)

    def verify(self, value: Any) -> bool:
        """
        Check whether the value can be assigned to this prop, e.g.
        min and max.

        Default implementation returns True.
        Override in subclass, if applicable.
        """
        return True

    def _get_value(self, frame: int) -> Any:
        """
        Returns the value, but not necessarily the correct type.

        IMPORTANT:
        Call ``self.value(frame)`` instead to convert to the prop's type and
        apply modifiers.
        """
        keys = sorted(self._keyframes)

        if len(keys) == 0:
            if self._value is None:
                if self.default is None:
                    if self.required:
                        raise ValueError("Both value and default are None.")
                    else:
                        return None
                return self.default
            return self._value

        elif len(keys) == 1:
            return keys[0].value

        else:
            if frame <= keys[0].frame:
                return keys[0].value
            elif frame >= keys[-1].frame:
                return keys[-1].value
            else:
                for i, k in enumerate(keys):
                    if k.frame == frame:
                        return k.value
                    if k.frame > frame:
                        break

                k1 = keys[i-1]
                k2 = keys[i]
                return interpolate(k1, k2, frame)

    def value(self, frame: int, use_mods: bool = True,
            default: Optional[Accessor] = None) -> Any:
        """
        Returns value at frame. Uses keyframe interpolations.
        Converts to type. Applies modifiers.
        """
        v = self.type(self._get_value(frame))
        if use_mods:
            for m in self.mods:
                v = m(default, v)
        return v


class BoolProp(Property):
    """
    Boolean.
    """
    type = bool
    supported_interps = {Interp.CONSTANT}


class IntProp(Property):
    """
    Integer.
    Min and max inclusive.
    Coords: Whether this quantity is in coords.
    """
    type = int
    supported_interps = {Interp.CONSTANT, Interp.LINEAR, Interp.QUADRATIC}

    min: int
    max: int
    coords: bool

    def __init__(self, min: Optional[int] = None, max: Optional[int] = None,
            coords: bool = False, **kwargs) -> None:
        self.min = min
        self.max = max
        super().__init__(**kwargs)

    def verify(self, value: int) -> bool:
        """
        Checks min and max.
        """
        if self.min is not None and value < self.min:
            return False
        if self.max is not None and value > self.max:
            return False
        return True


class FloatProp(Property):
    """
    Float.
    Min and max inclusive.
    Coords: Whether this quantity is in coords.
    """
    type = float
    supported_interps = {Interp.CONSTANT, Interp.LINEAR, Interp.QUADRATIC}

    min: int
    max: int
    coords: bool

    def __init__(self, min: Optional[float] = None, max: Optional[float] = None,
            coords: bool = False, **kwargs) -> None:
        self.min = min
        self.max = max
        super().__init__(**kwargs)

    def verify(self, value: float) -> bool:
        """
        Checks min and max.
        """
        return IntProp.verify(self, value)


class StrProp(Property):
    """
    String.
    Min and max inclusive.
    """
    type = str
    supported_interps = {Interp.CONSTANT}

    min: int
    max: int

    def __init__(self, min_len: Optional[int] = None,
            max_len: Optional[int] = None, **kwargs) -> None:
        self.min_len = min_len
        self.max_len = max_len
        super().__init__(**kwargs)

    def verify(self, value: str) -> bool:
        """
        Checks length min and max.
        """
        if self.min_len is not None and value < self.min_len:
            return False
        if self.max_len is not None and value > self.min_len:
            return False
        return True


class PathProp(StrProp):
    """
    Path property.
    Can verify if a path exists.
    """

    def __init__(self, isfile: bool = False, isdir: bool = False, **kwargs):
        """
        Initialize the property.

        :param isfile: Path must be a file.
        :param isdir: Path must be a directory.
        """
        self.isfile = isfile
        self.isdir = isdir
        super().__init__(**kwargs)

    def verify(self, value: str) -> bool:
        """
        Checks path isfile and isdir, if respective attributes are True.
        """
        if not super().verify(value):
            return False

        if self.isfile and not os.path.isfile(value):
            return False
        if self.isdir and not os.path.isdir(value):
            return False

        return True


class ArrayProp(Property):
    """
    Numpy array property.
    """
    type = np.array
    supported_interps = {Interp.CONSTANT, Interp.LINEAR, Interp.QUADRATIC}

    shape: Optional[Tuple[int]]

    def __init__(self, shape: Optional[Tuple[int]] = None, **kwargs) -> None:
        self.shape = shape
        super().__init__(**kwargs)

    def verify(self, value: np.ndarray) -> bool:
        """
        Checks shape.
        """
        if self.shape is not None and value.shape != tuple(self.shape):
            return False
        return True


class RGBProp(ArrayProp):
    """
    RGB color property, 0 to 255.
    """
    shape = (3,)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)


if __name__ == "__main__":
    b1 = BoolProp(name="asdf", desc="desc", default=False, animatable=False)
    b2 = BoolProp(name="asdf", desc="desc", default=False, animatable=True)

    try:
        b1.animate(Keyframe(0, False, Interp.CONSTANT))
    except AssertionError as e:
        print("Cannot animate b1:", e)

    print(b2.value(0))
    b2.animate(Keyframe(0, True, Interp.CONSTANT))
    print(b2.value(0), b2.value(1))
    b2.animate(Keyframe(5, False, Interp.CONSTANT))
    print(b2.value(-1), b2.value(0), b2.value(2), b2.value(5), b2.value(6))

    i = IntProp(name="asdf", desc="desc", default=10, animatable=True)
    print(i.value(0))
    i.animate(Keyframe(0, 5, Interp.LINEAR))
    print(i.value(0))
    i.animate(Keyframe(5, 10, Interp.CONSTANT))
    i.animate(Keyframe(10, 15, Interp.CONSTANT))
    print([i.value(a) for a in range(0, 6)])
    print([i.value(a) for a in range(5, 11)])



import os
import sys
from pathlib import Path

from .scene import Scene


def import_scene(file: str, name: str) -> Scene:
    """
    Import a scene object class from a Python file.
    """
    path = Path(file)
    sys.path.insert(0, str(path.absolute().parent))

    mod = __import__(path.with_suffix("").name)
    cls = getattr(mod, name)

    sys.path.pop(0)
    return cls



"""
Logging utils.
"""

import sys
import termcolor
from datetime import datetime


def time():
    now = datetime.now()
    return now.strftime("%H:%M:%S")


def log(type: str, msg: str, color: str):
    s = f"[{time()}] {type}:"
    s += " " * (6-len(type))
    s += msg
    print(termcolor.colored(s, color), file=sys.stderr)

def info(msg: str) -> None:
    """
    Info log to stderr.
    Color: blue
    """
    log("INFO", msg, "cyan")

def warn(msg: str) -> None:
    """
    Warning log to stderr.
    Color: yellow
    """
    log("WARN", msg, "yellow")

def error(msg: str) -> None:
    """
    Error log to stderr.
    Color: red
    """
    log("ERROR", msg, "red")



from .default import DefaultScene
from .keyframe import Keyframe, Interp
from .importer import import_scene
from .modifiers import Modifier, Coords
from .pgroup import PropertyGroup
from .props import *
from .scene import Scene



from typing import Any, Mapping


class Accessor:
    """
    Can access stuff with getattr.
    """
    _attrs: Mapping[str, Any]

    def __init__(self, attrs: Mapping[str, Any]):
        """
        All dicts in attrs are replaced with an Accessor object.
        """
        self._attrs = {}
        for k, v in attrs.items():
            if isinstance(v, dict):
                v = Accessor(v)
            self._attrs[k] = v

    def __getattr__(self, name: str) -> Any:
        return self._attrs[name]

    def _as_dict(self) -> Mapping[str, Any]:
        ret = {}
        for k, v in self._attrs.items():
            if isinstance(v, Accessor):
                v = v._as_dict()
            ret[k] = v

        return ret



"""
Default scene, contains settings for PianoRay.
i.e. default "implementation" of Scene.
"""

from .modifiers import *
from .pgroup import PropertyGroup
from .props import *
from .scene import Scene


class VideoProps(PropertyGroup):
    """
    Output video parameters.
    """

    resolution: ArrayProp(
        name="Resolution",
        desc="Output video resolution.",
        animatable=False,
        default=(1920, 1080),
        shape=(2,),
    )

    fps: IntProp(
        name="FPS",
        desc="Frames per second of output video.",
        animatable=False,
        default=30,
        min=1,
    )

    vcodec: StrProp(
        name="Video Codec",
        desc="Codec for video, passed to FFmpeg.",
        animatable=False,
        default="libx265",
    )


class AudioProps(PropertyGroup):
    """
    Audio.
    """

    file: PathProp(
        name="Audio File",
        desc="Path to audio file.",
        animatable=False,
        default=None,
        isfile=True,
    )

    start: FloatProp(
        name="Start Time",
        desc="Timestamp, in seconds, you press the first note.",
        animatable=False,
        default=0,
    )


class CompositionProps(PropertyGroup):
    """
    Composition, i.e. structure of the video.
    """

    margin_start: FloatProp(
        name="Start Margin",
        desc="Pause, in seconds, before first note starts.",
        animatable=False,
        default=3,
        min=0,
    )

    margin_end: FloatProp(
        name="End Margin",
        desc="Pause, in seconds, after the last note ends.",
        animatable=False,
        default=3,
        min=0,
    )

    fade_in: FloatProp(
        name="Fade In",
        desc="Seconds of fade in.",
        animatable=False,
        default=1,
        min=0,
    )

    fade_out: FloatProp(
        name="Fade Out",
        desc="Seconds of fade out.",
        animatable=False,
        default=1,
        min=0,
    )

    fade_blur: FloatProp(
        name="Fade Blur",
        desc="Blur radius of fade in coords.",
        animatable=False,
        mods=[Coords()],
        default=1,
        coords=True,
    )


class PianoProps(PropertyGroup):
    """
    Piano parameters.
    """

    black_width_fac: FloatProp(
        name="Black Key Width Factor",
        desc="Black key width as factor of white key width.",
        animatable=False,
        default=0.6,
        min=0,
    )


class BlocksProps(PropertyGroup):
    """
    The blocks that fall down.
    """

    speed: FloatProp(
        name="Speed",
        desc="If ``X`` is the distance between the top of the screen and the"
             "top of the keyboard, the blocks travel ``speed * X`` per second.",
        default=0.5,
    )

    color: RGBProp(
        name="Color",
        desc="Color of the blocks.",
        default=(150, 160, 240),
    )

    radius: FloatProp(
        name="Corner Radius",
        desc="Corner rounding radius in coords.",
        mods=[Coords()],
        default=0.25,
        min=0,
    )

    glow_intensity: FloatProp(
        name="Glow Intensity",
        desc="Intensity of glow around blocks.",
        default=0.3,
        min=0,
    )

    glow_color: RGBProp(
        name="Glow Color",
        desc="Color of the glow.",
        default=(230, 230, 255),
    )

    glow_radius: FloatProp(
        name="Glow Radius",
        desc="Radius of glow around blocks in coords.",
        mods=[Coords()],
        default=0.4,
        coords=True,
    )


class MidiProps(PropertyGroup):
    file: PathProp(
        name="MIDI File",
        desc="Path to MIDI file.",
        animatable=False,
        default=None,
        isfile=True,
    )

    speed: FloatProp(
        name="Speed Multiplier",
        desc="MIDI notes speed multiplier.",
        animatable=False,
        default=1,
    )

    min_length: FloatProp(
        name="Minimum Duration",
        desc="Min duration of a note in seconds.",
        animatable=False,
        default=0.08,
        min=0,
    )


class KeyboardProps(PropertyGroup):
    file: PathProp(
        name="Video File",
        desc="Path to video recording of keyboard.",
        animatable=False,
        default=None,
        isfile=True,
    )

    start: FloatProp(
        name="Start",
        desc="Timestamp, in seconds, when the first note starts in the video.",
        animatable=False,
        default=None,
    )

    end: FloatProp(
        name="End",
        desc="Timestamp, in seconds, when the last note starts in the video.",
        animatable=False,
        default=None,
    )

    crop: ArrayProp(
        name="Crop",
        desc="Crop points of the keyboard. See docs for more info.",
        animatable=False,
        default=None,
        shape=(4, 2),
    )

    dim_mult: FloatProp(
        name="Multiplicative Dimming",
        desc="Multiplier to pixel brightness.",
        default=1,
        min=0,
    )

    dim_add: FloatProp(
        name="Additive Dimming",
        desc="Value added to pixel brightness (0 to 255).",
        default=0,
    )

    below_length: FloatProp(
        name="Length of Below Section",
        desc="Length in coords of section below keyboard.",
        mods=[Coords()],
        default=7,
        min=0,
    )

    octave_lines: BoolProp(
        name="Octave Lines",
        desc="Whether to render octave lines.",
        default=True,
    )


class GlareProps(PropertyGroup):
    radius: FloatProp(
        name="Radius",
        desc="Radius of glare in coords.",
        mods=[Coords()],
        default=3,
        min=0,
    )

    intensity: FloatProp(
        name="Intensity",
        desc="Intensity of glare.",
        default=0.9,
        min=0,
    )

    jitter: FloatProp(
        name="Jitter",
        desc="Range of random multiplier.",
        default=0.08,
        min=0,
    )

    streaks: IntProp(
        name="Streaks",
        desc="Number of streaks.",
        default=6,
        min=0,
        max=20,  # C implementation limitation.
    )


class ParticleProps(PropertyGroup):
    pps: FloatProp(
        name="Particles per Second",
        desc="Number of particles to emit per note per second.",
        default=40,
        min=0,
    )

    air_resist: FloatProp(
        name="Air Resistance",
        desc="Velocity multiplies by this every second.",
        default=0.6,
        min=0,
    )

    lifetime: FloatProp(
        name="Lifetime",
        desc="Particle lifetime in seconds.",
        default=3,
        min=0,
    )

    x_vel: FloatProp(
        name="X Velocity",
        desc="Initial X velocity range in coords/sec.",
        mods=[Coords(), SecToFrame()],
        default=1,
        min=0,
    )

    y_vel: FloatProp(
        name="Y Velocity",
        desc="Initial Y velocity range in coords/sec.",
        mods=[Coords(), SecToFrame()],
        default=4,
        min=0,
    )

    wind_strength: FloatProp(
        name="Wind Strength",
        desc="Strength multiplier of wind affecting particles.",
        default=1,
    )

    heat_strength: FloatProp(
        name="Heat Strength",
        desc="Strength multiplier of heat affecting particles.",
        default=1,
    )

    gravity: FloatProp(
        name="Gravity",
        desc="Strength multiplier of gravity affecting particles.",
        default=1,
    )


class DefaultScene(Scene):
    _pgroups = {
        "audio": AudioProps(),
        "blocks": BlocksProps(),
        "composition": CompositionProps(),
        "glare": GlareProps(),
        "keyboard": KeyboardProps(),
        "midi": MidiProps(),
        "ptcls": ParticleProps(),
        "piano": PianoProps(),
        "video": VideoProps(),
    }



from typing import Tuple

import cv2
import numpy as np

from ..utils import interp
from .effect import Effect


class VideoRead:
    """
    Read frames of a video.
    Maps video frames to client frames.
    """

    def __init__(self, path: str, fps: int, mapping: Tuple[int, int, int, int]):
        """
        Initialize.

        :param path: Path of video.
        :param fps: Client fps.
        :param mapping: Tuple of
            ``(client_start, client_end, src_start, src_end)`` frames.
        """
        self._video = cv2.VideoCapture(path)
        self.mapping = mapping

        # _real_frame stores frame number with first frame of video = 0
        self._real_frame = 0
        self._last = None  # Last read img

        self._read_next()

    def _get_frame(self, frame: int) -> int:
        """
        Return frame of input video corresponding to client's frame.
        """
        f = interp(frame, self.mapping[:2], self.mapping[2:])
        return round(f)

    def _read_next(self):
        """
        Read next frame, store in self._last, and
        increment self._frame.
        """
        ret, img = self._video.read()
        self._real_frame += 1
        if ret:
            self._last = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    def read(self, frame: int) -> np.ndarray:
        """
        Read frame. Pass the frame the client needs.
        Can only read monotonically.
        """
        f = self._get_frame(frame)
        while self._real_frame < f:
            self._read_next()

        return self._last


class Keyboard(Effect):
    """
    Piano keyboard rendering.
    """

    def __init__(self, props, cache, libs, notes) -> None:
        """
        :param notes: Parsed MIDI notes.
        """
        super().__init__(props, cache, libs)

        fps = props.video.fps
        duration = notes[-1].start - notes[0].start
        self.video = VideoRead(props.keyboard.file, fps,
            (0, duration, props.keyboard.start*fps, props.keyboard.end*fps))

        self.compute_crop(props)

    @staticmethod
    def extend_vec(v1, v2, length):
        """
        Let u be the unit vector in the direction from v1 to v2.
        Let v be u * length.
        Returns v + v2
        """
        diff = v2 - v1
        u = diff / np.linalg.norm(diff)
        v = u * length
        return v + v2

    def compute_crop(self, props):
        crop = np.array(props.keyboard.crop)
        src_width = np.linalg.norm(crop[1]-crop[0])
        dst_width = props.video.resolution[0]
        scale = src_width / dst_width

        below_len = props.keyboard.below_length * scale
        src_points = np.array(
            (
                crop[0],
                crop[1],
                self.extend_vec(crop[1], crop[2], below_len),
                self.extend_vec(crop[0], crop[3], below_len),
            ),
            dtype=np.float32)

        dst_width = props.video.resolution[0]
        dst_height = np.linalg.norm(src_points[3]-src_points[0]) / scale
        dst_kbd_height = np.linalg.norm(crop[3]-crop[0]) / scale
        dst_width, dst_height, dst_kbd_height = \
            map(int, (dst_width, dst_height, dst_kbd_height))
        dst_points = np.array(
            ((0,0), (dst_width,0), (dst_width,dst_height), (0,dst_height)),
            dtype=np.float32)

        mask = np.empty((dst_height, dst_width, 3), dtype=np.float32)
        for y in range(dst_height):
            if y < dst_kbd_height:
                fac = 1
            else:
                fac = np.interp(y, (dst_kbd_height, dst_height), (0, 1))
                fac **= 2
                fac = np.interp(fac, (0, 1), (1, 0.4))

            mask[y, :] = fac

        self.persp = cv2.getPerspectiveTransform(src_points, dst_points)
        self.dst_shape = (dst_width, dst_height)
        self.mask = mask

    def render(self, props, img: np.ndarray, frame: int):
        """
        Render the keyboard.
        """
        dst = self.dst_shape

        kbd = self.video.read(frame)
        kbd = cv2.warpPerspective(kbd, self.persp, dst).astype(np.float64)
        kbd *= self.mask

        kbd *= props.keyboard.dim_mult
        kbd += props.keyboard.dim_add
        kbd[kbd<0] = 0
        kbd[kbd>255] = 255
        kbd = kbd.astype(np.uint8)

        half = int(props.video.resolution[1] / 2)
        img[half:half+dst[1], 0:dst[0], ...] = kbd

        # Octave lines
        if props.keyboard.octave_lines:
            self.libs["keyboard"].render_octave_lines(
                img, img.shape[1], img.shape[0])



from typing import Any, Mapping, Optional

from .accessor import Accessor
from .props import Property


class PropertyGroup:
    """
    Group of properties. Define a subclass to create your PropertyGroup.
    Define properties by creating annotations with ``:``. Don't override
    any methods, as instancing a PropertyGroup subclass requires the
    methods.

    .. code-block:: py

        class MyProps(PropertyGroup):
            temperature: FloatProp(
                name="Temperature",
                desc="Temperature to cook the food at.",
                default=-10,
            )

            food: StringProp(
                name="Food",
                desc="The food to cook.",
                default="Java",
            )

    You can set and get properties.

    .. code-block:: py

        pgroup.temperature                # Returns the property object.
        pgroup.temperature.animate(...)   # Animate. See Property docs.
        pgroup.temperature = -273         # Calls pgroup.temperature.set_value()
    """

    _props: Mapping[str, Property]

    def __init__(self):
        """
        Reads __annotations__ and stores in ``self._props``.
        """
        object.__setattr__(self, "_props", {})

        for k, v in self.__annotations__.items():
            if isinstance(v, Property):
                self._props[k] = v

    def __setattr__(self, name: str, value: Any):
        """
        ``pgroup.prop_name = 1``
        is equivalent to
        ``pgroup.prop_name.set_value(1)``
        """
        self._props[name].set_value(value)

    def __getattr__(self, name: str) -> Property:
        return self._props[name]

    def _values(self, frame: int, use_mods: bool = True,
            default: Optional[Accessor] = None) -> Mapping[str, Any]:
        """
        Get values of all properties at frame.
        Returns ``{"prop_name": value}``.
        """
        ret = {}
        for k, prop in self._props.items():
            v = prop.value(frame, use_mods, default)
            ret[k] = v

        return ret



from typing import Any, Mapping, Type

from .accessor import Accessor
from .pgroup import PropertyGroup


class Scene:
    """
    Group of PropertyGroups.

    Create a subclass with your pgroups. Set the dictionary ``_pgroups`` to
    mapping of id to property group instance.

    .. code-block:: py

        class MyScene(Scene):
            _pgroups = {
                "food": FoodProps(),
            }

    Create a subclass of a scene, and override the setup method to do animation.
    Scene.setup is called at initialize time.

    .. code-block:: py

        # We are extending "MyScene", described above.
        class MyOtherScene(MyScene):
            def setup(self):
                self.food.temperature = 100
    """
    _pgroups: Mapping[str, PropertyGroup]

    def __init__(self):
        """
        Initializes pgroups.
        """
        self.setup()

    def __getattr__(self, name: str) -> PropertyGroup:
        if name == "default":
            return object.__getattribute__(self, "default")
        return object.__getattribute__(self, "_pgroups")[name]

    def values(self, frame: int, use_mods: bool = True) -> Accessor:
        """
        Returns Accesor object of all pgroup values at frame.
        """
        default = self.values(0, False) if use_mods else None

        ret = {}
        for k, pgroup in self._pgroups.items():
            v = pgroup._values(frame, use_mods, default)
            ret[k] = v

        return Accessor(ret)

    @property
    def default(self) -> Accessor:
        """
        Equivalent to ``self.values(0)``.
        Usually used to get non animatable props.
        """
        return self.values(0)

    def setup(self) -> None:
        """
        Do any animation or property value setting here.
        """



from typing import Any

import numpy as np

from .keyframe import Keyframe, Interp


def lerp(fac, from_range, to_range) -> Any:
    """
    Linear interpolation, similar to np.interp
    Works on np.array as well.
    """
    progress = (fac-from_range[0]) / (from_range[1]-from_range[0])
    return to_range[0] + progress * (to_range[1]-to_range[0])


def quad_fac(fac):
    """
    if fac < 0.5: y = 2x^2
    else: y = -2(x-1)^2 + 1
    """
    if fac < 0.5:
        return 2 * fac**2
    else:
        return -2 * (fac-1)**2 + 1

def interpolate(k1: Keyframe, k2: Keyframe, frame: int) -> Any:
    """
    Interpolate between two keyframes.
    k1.frame <= frame < k2.frame
    """
    interp = k1.interp
    v1 = k1.value
    v2 = k2.value
    f1 = k1.frame
    f2 = k2.frame

    if interp == Interp.CONSTANT:
        return v1

    elif interp in (Interp.LINEAR, Interp.QUADRATIC):
        fac = np.interp(frame, (f1, f2), (0, 1))  # Lerp
        if interp == Interp.QUADRATIC:
            fac = quad_fac(fac)

        return lerp(fac, (0, 1), (v1, v2))



"""
PianoRay module.
"""

import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
del os

from . import effects
from . import render
from . import utils

from .api import *

__version__ = utils.VERSION



import pygame

from .timeline import Timeline
from .utils import *
from .video import Video

pygame.init()


class Info:
    """
    Shows info as text to user.
    """

    def __init__(self, video: Video, timeline: Timeline):
        self.video = video
        self.timeline = timeline

    def draw(self, surface: pygame.Surface, events, rect):
        """
        Draw info on surface.

        :param rect: ``(x, y, w, h)`` coordinates to draw.
        """
        frame = self.timeline.frame
        fps = self.video.fps
        text = [
            f"Frame: {frame}",
            f"Cached frames: {self.video.extracted}",
            f"Timestamp: {frame/fps:.3f}s",
        ]

        x = rect[0] + 20
        for i, t in enumerate(text):
            y = rect[1] + 10 + 20*i
            surf = FONT_MED.render(t, True, WHITE)
            surface.blit(surf, (x, y))



"""
C++ library compilation and handling.
"""

import ctypes
import os
from pathlib import Path
from subprocess import Popen
from typing import Sequence

import numpy as np
from numpy.ctypeslib import ndpointer

from . import logger
from .utils import GCC

ROOT = Path(__file__).absolute().parent

CPP_UTILS = ROOT / "cutils"


class Types:
    """
    C++ types. Assign a sequence of types to func.argtypes:

    lib.render_thing.argtypes = [img, int, int, float, double, ...]
    """
    _arr_flags = "aligned, c_contiguous"

    char = ctypes.c_int8
    uchar = ctypes.c_uint8
    int = ctypes.c_int32
    uint = ctypes.c_uint32
    float = ctypes.c_float
    double = ctypes.c_double

    arr_uchar = ndpointer(dtype=uchar, ndim=1, flags=_arr_flags)
    arr_int = ndpointer(dtype=int, ndim=1, flags=_arr_flags)
    arr_double = ndpointer(dtype=double, ndim=1, flags=_arr_flags)

    img = ndpointer(dtype=uchar, ndim=3, flags=_arr_flags)
    path = ndpointer(dtype=char, ndim=1, flags=_arr_flags)

    @staticmethod
    def cpath(path):
        """
        Return C path (numpy array of chars, null terminated).
        """
        data = [c for c in str(path).encode()]
        data.append(0)
        return np.array(data, dtype=np.int8)


def build_lib(files: Sequence[str], cache: Path, name: str) -> ctypes.CDLL:
    """
    Build and load a library.

    :param files: C files relative to THIS file.
    :param cache: Cache directory.
    :param name: Name of the library.
    :return: C library.
    """
    logger.info(f"Building C library {name}")

    cache = cache / name
    os.makedirs(cache, exist_ok=True)

    files = [ROOT/f for f in files]

    obj_files = []
    for f in files:
        obj_path = str(cache / f.with_suffix(".o").name)
        obj_files.append(obj_path)
        compile(str(f), obj_path)

    lib_path = str(cache / f"lib{name}.so")
    link(obj_files, lib_path)

    return ctypes.CDLL(lib_path)

def compile(cpp, obj):
    """
    Compile a C++ file and output to obj file.
    """
    args = [GCC, "-Wall", "-O3", "-c", "-fPIC", cpp, "-o", obj, "-I", CPP_UTILS]
    p = Popen(args)
    p.wait()
    assert p.returncode == 0

def link(obj_files, lib_path):
    """
    Link object files.
    """
    args = [GCC, "-shared", "-o", lib_path, *obj_files]
    p = Popen(args)
    p.wait()
    assert p.returncode == 0
